#include <avr/io.h>
#include "lcd-routines.h"
#include "wait1000ms.h"
#include "config.h"
#include <util/delay.h>
#include <avr/sleep.h>
#include <stdlib.h>
#include <string.h>
#include <avr/eeprom.h>
#include <avr/pgmspace.h>
#include <avr/wdt.h>
#include <math.h>
//#define DebugOut 3		// wenn gesetzt, Zwischenausgaben bei Widerstandmessung in Zeile 2,3,4
//#define DebugOut 4		// wenn gesetzt, Spannungsausgabe bei Diodenmessung in Zeilen 3+4
//#define DebugOut 5		// wenn gesetzt, Zwischenausgaben der TransistorUntersuchung in Zeile 2+3
//#define DebugOut 10		// wenn gesetzt, Zwischenausgaben in ReadCapacity in Zeilen 3+4 
#include "Transistortester.h"

//Programmbeginn
int main(void) {
  //Einschalten
  ON_DDR = (1<<ON_PIN);
  ON_PORT = (1<<ON_PIN) | (1<<RST_PIN);	//Strom an und Pullup fr Reset-Pin
  uint8_t tmp;
  //ADC-Init
  ADCSRA = (1<<ADEN) | AUTO_CLOCK_DIV;	//Vorteiler=8
  lcd_init();
	
  wdt_disable();
  if(MCU_STATUS_REG & (1<<WDRF)) { 
     // berprfen auf Watchdog-Reset 
     // Das tritt ein, wenn der Watchdog 2s nicht zurckgesetzt wurde
     // Kann vorkommen, wenn sich das Programm in einer Endlosschleife "verheddert" hat.
     lcd_fix_string(TestTimedOut);	//Timeout-Meldung
     wait3s();				//warte 3 Sekunden
     ON_PORT = 0;			//Abschalten!
     return 0;
  }
  LCDLoadCustomChar(LCD_CHAR_DIODE1);	//Custom-Zeichen
  lcd_fix_customchar(DiodeIcon1);	//lade Zeichen  >|
  LCDLoadCustomChar(LCD_CHAR_DIODE2);	//Custom-Zeichen 
  lcd_fix_customchar(DiodeIcon2);	//lade Zeichen  |<
#ifdef C_MESS
  LCDLoadCustomChar(LCD_CHAR_CAP);	//Custom-Zeichen 
  lcd_fix_customchar(CapIcon);		//lade Zeichen  ||
#endif
#ifdef R_MESS
  LCDLoadCustomChar(LCD_CHAR_RESIS1);	//Custom-Zeichen 
  lcd_fix_customchar(ResIcon1);		//lade Zeichen  [
  LCDLoadCustomChar(LCD_CHAR_RESIS2);	//Custom-Zeichen 
  lcd_fix_customchar(ResIcon2);		//lade Zeichen  ]
#endif
  
#ifdef LCD_CYRILLIC
  //bei kyrillischen LCD-Zeichensatz Omega- und -Zeichen laden
  LCDLoadCustomChar(LCD_CHAR_OMEGA);	//Custom-Zeichen
  //Omega-Zeichen in LCD laden
  lcd_fix_customchar(CyrillicOmegaIcon);
  LCDLoadCustomChar(LCD_CHAR_U);	//Custom-Zeichen
  //-Zeichen in LCD laden
  lcd_fix_customchar(CyrillicMuIcon);
#endif

//*****************************************************************
//Einsprungspunkt, wenn Start-Taste im Betrieb erneut gedrckt wird
start:
  PartFound = PART_NONE;
#ifdef R_MESS
  NumOfR = 0;
#endif
  NumOfDiodes = 0;
  PartReady = 0;
  PartMode = 0;
  lcd_clear();
  ADC_DDR = TXD_MSK;		//Software-UART aktivieren
#ifdef C_MESS
  ca = 0;
  cb = 0;
#endif
#ifdef PULLUP_DISABLE
  MCUCR = (1<<PUD);		// disable Pull-Up Resistors
#endif
  uart_newline();
  ref_mv = ReadADC(0x0e);      // Referenz-Spannung lesen
  ref_mv = W5msReadADC(0x0e);  // Referenz-Spannung lesen
#ifdef AUTOSCALE_ADC
   mindiff = ref_mv;
   for (multip=2;multip<64;multip++) {
      // Umrechnungsfaktoren von ADC-Auflsung in mV finden
      //  ADC * multip / divid 
      for (divid=2;divid<multip;divid++) {
         ergeb = (unsigned int)(1023 * multip) / divid;
#ifdef __AVR_ATmega8__
         diff = abs((int)((ref_mv*2) - ergeb));
#else
         diff = abs((int)(ref_mv - ergeb));
#endif
#if ANZ_MES == 44
         if (!(diff > mindiff)) 
            // so viel wie mglich addieren, grsste minmul,mindiv
#else
         if (diff < mindiff) 
            // so schnell wie mglich zum Ergebnis, kleinste minmul,mindiv
#endif
                            {
            mindiff = diff;
            minmul = multip;
            mindiv = divid;
         }
      }
   }
   // durch minmul und mindif wird die Auflsung des ADC in der Funktion ReadADC in mV umgerechnet,
   // wenn auf interne Referenz umgeschaltet wird.

#if 0
   Line2();
   lcd_string(utoa(ref_mv, outval, 10));
   lcd_data(':');
   lcd_string(utoa(minmul, outval, 10));
   lcd_data('/');
   lcd_string(utoa(mindiv, outval, 10));
   wait2s();
#endif
#endif
  ref_mv += REF_KORR;		// Korrektur fr Kondensator-Messung

  ReadADC(5);	//Dummy-Readout
  trans.hfe[0] = W5msReadADC(5); 	//mit 5V Referenz
    				// 5001 = 2,56V, durch Teiler 10,317V
  Line1();	//1. Zeile
  lcd_fix_string(Bat);		//Anzeige: "Bat. "
#ifdef BAT_OUT
  utoa(trans.hfe[0]/((100*33)/133), outval, 10);	// nur 2 Stellen ausgeben
  lcd_data(outval[0]);
  lcd_data('.');
  lcd_data(outval[1]);
  lcd_data('V');
  lcd_data(' ');
#endif
#ifdef WDT_enabled
  wdt_enable(WDTO_2S);		//Watchdog an
#endif
  if (trans.hfe[0] <  (((unsigned long)7300*(unsigned long)33)/133)) {
     //Vcc < 7,3V; Warnung anzeigen
     if(trans.hfe[0] < (((unsigned long)6300*(unsigned long)33)/133)) {	
        //Vcc <6,3V; zuverlssiger Betrieb nicht mehr mglich
        lcd_fix_string(BatEmpty);	//Batterie leer!
        wait2s();
        PORTD = 0;			//abschalten
        return 0;
     }
     lcd_fix_string(BatWeak);		//Batterie schwach
  } else { // Batterie-Spannung ausreichend
     lcd_fix_string(OK); 		// "OK"
  }

  wait1s();			// warten, damit man Batteriespannung ablesen kann
  //Test beginnen
#ifdef WITH_AUTO_REF
  RefVoltage();			//kompensiere Referenz-Spannung
#endif
  Line2();
  lcd_fix_string(TestRunning);		//String: Test luft
#ifndef DebugOut
  Line2();
#endif
#ifdef C_MESS
  EntladePins();		// evtl. vorhandene Restladung entladen
#endif
#ifdef WITH_SELFTEST
  AutoCheck();			//prfe, ob Selbst-Prfung durchgefhrt werden soll
#endif
     
  //Alle 6 Kombinationsmglichkeiten fr die 3 Pins prfen
  CheckPins(TP1, TP2, TP3);
  CheckPins(TP1, TP3, TP2);
  CheckPins(TP2, TP1, TP3);
  CheckPins(TP2, TP3, TP1);
  CheckPins(TP3, TP2, TP1);
  CheckPins(TP3, TP1, TP2);
  
#ifdef C_MESS
 #if (CapTestMode & 0x30) > 0
  //Separate Messung zum Test auf Kondensator
  if(((PartFound == PART_NONE) || (PartFound == PART_RESISTOR) || (PartFound == PART_DIODE)) ) {
     //Kondensator entladen; sonst ist evtl. keine Messung mglich
     EntladePins();		// evtl. vorhandene Restladung entladen
  #if (CapTestMode & 0x30) == 0x10
     // nur mit 2 Pins Kapazitten messen
     ReadCapacity((CapTestMode & 0x03),  (CapTestMode & 0x0c)>>2 );
  #else
     //Kapazitt in allen 6 (3) Pin-Kombinationen messen
     ReadCapacity(TP3, TP1);
     ReadCapacity(TP3, TP2);
//     ReadCapacity(TP2, TP3);
     ReadCapacity(TP2, TP1);
//     ReadCapacity(TP1, TP3);
//     ReadCapacity(TP1, TP2);
  #endif
  }
 #endif
#endif
  //Fertig, jetzt folgt die Auswertung
  lcd_clear();
  Line1();
  if(PartFound == PART_DIODE) {
     if(NumOfDiodes == 1) {		//Standard-Diode
        lcd_fix_string(Diode);		//"Diode: "
        lcd_ziff1(diodes[0].Anode);
        lcd_fix_string(AnKat);		//"->|-"
        lcd_ziff1(diodes[0].Cathode);
        UfAusgabe(0x70);
#ifdef C_MESS
        lcd_fix_string(GateCap);		//" C="
        ReadCapacity(diodes[0].Cathode,diodes[0].Anode);	// Kapazitt in Sperr-Richtung
        lcd_show_format_cap();
#endif
        goto end;
       }
     else if(NumOfDiodes == 2) { //Doppeldiode
        lcd_data('2');
        lcd_fix_string(Dioden);		//"Dioden "
        if(diodes[0].Anode == diodes[1].Anode) { //Common Anode
           lcd_ziff1(diodes[0].Cathode);
           lcd_fix_string(KatAn);	//"-|<-"
           lcd_ziff1(diodes[0].Anode);
           lcd_fix_string(AnKat);	//"->|-"
           lcd_ziff1(diodes[1].Cathode);
           UfAusgabe(0x01);
           goto end;
        } else if(diodes[0].Cathode == diodes[1].Cathode) { //Common Cathode
           lcd_ziff1(diodes[0].Anode);
           lcd_fix_string(AnKat);	//"->|-"
	   lcd_ziff1(diodes[0].Cathode);
           lcd_fix_string(KatAn);	//"-|<-"
           lcd_ziff1(diodes[1].Anode);
           UfAusgabe(0x01);
           goto end;
          }
        else if ((diodes[0].Cathode == diodes[1].Anode) && (diodes[1].Cathode == diodes[0].Anode)) {
           //Antiparallel
           lcd_ziff1(diodes[0].Anode);
           lcd_fix_string(AnKat);	//"->|-"
           lcd_ziff1(diodes[0].Cathode);
           lcd_fix_string(AnKat);	//"->|-"
           lcd_ziff1(diodes[1].Cathode);
           UfAusgabe(0x01);
//           lcd_fix_string(Antiparallel);	//"Antiparallel"
           goto end;
        }
     } else if(NumOfDiodes == 3) {
        //Serienschaltung aus 2 Dioden; wird als 3 Dioden erkannt
        trans.b = 3;
        trans.c = 3;
        /* berprfen auf eine fr eine Serienschaltung von 2 Dioden mgliche Konstellation
          Es stimmt nur einmal die Pin-Nr. einer Kathode mit einer Anode berein.
          Das kommmt daher, dass die Dioden als 2 Einzeldioden und ZUSTZLICH als eine "groe" Diode erkannt werden.
        */
        if(diodes[0].Cathode == diodes[1].Anode)
          {
           trans.b = 0;
           trans.c = 1;
          }
        if(diodes[0].Anode == diodes[1].Cathode)
          {
           trans.b = 1;
           trans.c = 0;
          }
        if(diodes[0].Cathode == diodes[2].Anode)
          {
           trans.b = 0;
           trans.c = 2;
          }
        if(diodes[0].Anode == diodes[2].Cathode)
          {
           trans.b = 2;
           trans.c = 0;
          }
        if(diodes[1].Cathode == diodes[2].Anode)
          {
           trans.b = 1;
           trans.c = 2;
          }
        if(diodes[1].Anode == diodes[2].Cathode)
          {
           trans.b = 2;
           trans.c = 1;
          }
#if DebugOut == 4
	Line3();
        lcd_ziff1(diodes[0].Anode);
        lcd_data(':');
        lcd_ziff1(diodes[0].Cathode);
        lcd_data(' ');
        lcd_string(utoa(diodes[0].Voltage, outval, 10));
        lcd_data(' ');
        lcd_ziff1(diodes[1].Anode);
        lcd_data(':');
        lcd_ziff1(diodes[1].Cathode);
        lcd_data(' ');
        lcd_string(utoa(diodes[1].Voltage, outval, 10));
	Line4();
        lcd_ziff1(diodes[2].Anode);
        lcd_data(':');
        lcd_ziff1(diodes[2].Cathode);
        lcd_data(' ');
        lcd_string(utoa(diodes[2].Voltage, outval, 10));
        Line1();
#endif
        if((trans.b<3) && (trans.c<3)) {
           lcd_data('3');
           lcd_fix_string(Dioden);	//"Dioden "
           lcd_ziff1(diodes[trans.b].Anode);
           lcd_fix_string(AnKat);	//"->|-"
           lcd_ziff1(diodes[trans.b].Cathode);
           lcd_fix_string(AnKat);	//"->|-"
           lcd_ziff1(diodes[trans.c].Cathode);
           UfAusgabe( (trans.b<<4)|trans.c);
           goto end;
        }
     }
     // end (PartFound == PART_DIODE)
  } else if (PartFound == PART_TRANSISTOR) {
    if(PartReady == 0) {	//Wenn 2. Prfung nie gemacht, z.B. bei Transistor mit Schutzdiode
       trans.hfe[1] = trans.hfe[0];
    }
    if((trans.hfe[0]>trans.hfe[1])) {
       //Wenn der Verstrkungsfaktor beim ersten Test hher war: C und E vertauschen!
       trans.hfe[1] = trans.hfe[0];
       tmp = trans.c;
       trans.c = trans.e;
       trans.e = tmp;
    }

    if(PartMode == PART_MODE_NPN) {
       lcd_fix_string(NPN);		//"NPN "
    } else {
       lcd_fix_string(PNP);		//"PNP "
    }
    if(NumOfDiodes > 2) {	//Transistor mit Schutzdiode
       if(PartMode == PART_MODE_NPN) {
          lcd_fix_string(AnKat);	//"->|-"
       } else {
          lcd_fix_string(KatAn);	//"-|<-"
       }
    }
    lcd_fix_string(ebcstr);		//" EBC="
    lcd_ziff1(trans.e);
    lcd_ziff1(trans.b);
    lcd_ziff1(trans.c);
    Line2(); //2. Zeile
    lcd_fix_string(hfestr);		//"hFE="
    lcd_string(utoa(trans.hfe[1], outval, 10));
    lcd_data(' ');

    for(trans.c=0;trans.c<NumOfDiodes;trans.c++) {
       if(((diodes[trans.c].Cathode == trans.e) && (diodes[trans.c].Anode == trans.b) && (PartMode == PART_MODE_NPN)) ||
          ((diodes[trans.c].Anode == trans.e) && (diodes[trans.c].Cathode == trans.b) && (PartMode == PART_MODE_PNP))) {
          lcd_fix_string(Uf);		//"Uf="
          lcd_string(utoa(diodes[trans.c].Voltage, outval, 10));
          lcd_data('m');
          lcd_data('V');
          goto end;
       }
    }
    goto end;
    // end (PartFound == PART_TRANSISTOR)
  } else if (PartFound == PART_FET) {	//JFET oder MOSFET
    if(PartMode&1) {
       //N-Kanal
       lcd_data('N');
    } else {
       lcd_data('P');			//P-Kanal
    }
    lcd_data('-');
    if((PartMode==PART_MODE_N_D_MOS) || (PartMode==PART_MODE_P_D_MOS)) {
       lcd_data('D');			// N-D
       lcd_fix_string(mosfet);		//"-MOS"
    } else {
       if((PartMode==PART_MODE_N_JFET) || (PartMode==PART_MODE_P_JFET)) {
          lcd_fix_string(jfet);		//"-JFET"
       } else {
          lcd_data('E');		// -E
          lcd_fix_string(mosfet);	//"-MOS"
       }
    }
 #ifdef C_MESS	//Gatekapazitt
    if(PartMode < 3) {  //Anreicherungs-MOSFET
       lcd_fix_string(GateCap);		//" C="
       ReadCapacity(trans.b,trans.e);	//Messung Kapazitt
       lcd_show_format_cap();
    }
 #endif
    Line2(); //2. Zeile
    lcd_fix_string(gds);		//"GDS="
    lcd_ziff1(trans.b);
    lcd_ziff1(trans.c);
    lcd_ziff1(trans.e);
    if((NumOfDiodes > 0) && (PartMode < 3)) {
       //MOSFET mit Schutzdiode; gibt es nur bei Anreicherungs-FETs
       lcd_data(LCD_CHAR_DIODE1);	//Diode anzeigen >|
    } else {
       lcd_data(' ');	//Leerzeichen
    }
    if(PartMode < 3) {	//Anreicherungs-MOSFET
       lcd_fix_string(vt);
       lcd_string(utoa(gthvoltage, outval, 10));	//Gate-Schwellspannung, wurde zuvor ermittelt
       lcd_data('m');
    }
    goto end;
    // end (PartFound == PART_FET)
  } else if (PartFound == PART_THYRISTOR) {
    lcd_fix_string(Thyristor);		//"Thyristor"
    goto gakAusgabe;
  } else if (PartFound == PART_TRIAC) {
    lcd_fix_string(Triac);		//"Triac"
    goto gakAusgabe;
  }
 #ifdef R_MESS	//Widerstandsmessung nur mit Mega8 verfgbar
  else if(PartFound == PART_RESISTOR) {
//    lcd_fix_string(Resistor); 		//"Widerst: "
    if (NumOfR == 1) { // Einzelwiderstand
       lcd_ziff1(resis[ii].rb);  	//Pin-Angabe
       lcd_fix_string(Resis);
       lcd_ziff1(resis[ii].ra);		//Pin-Angaben
    } else { // R-Max suchen
       ii = 0;
       if (resis[1].rx > resis[0].rx)
          ii = 1;
       if (NumOfR == 2) {
          ii = 2;
       } else {
          if (resis[2].rx > resis[ii].rx) {
             ii = 2;
          }
       }

       if (ii == 0) {
          lcd_data('1');
          lcd_fix_string(Resis);	// -[=]-
          lcd_data('3');
          lcd_fix_string(Resis);	// -[=]-
          lcd_data('2');
       }
       if (ii == 1) {
          lcd_data('1');
          lcd_fix_string(Resis);	// -[=]-
          lcd_data('2');
          lcd_fix_string(Resis);	// -[=]-
          lcd_data('3');
       }
       if (ii == 2) {
          lcd_data('2');
          lcd_fix_string(Resis);	// -[=]-
          lcd_data('1');
          lcd_fix_string(Resis);	// -[=]-
          lcd_data('3');
       }
    }
    Line2(); //2. Zeile
    if (NumOfR == 1) {
       RvalOut(0);
    } else {
       if (ii == 0) {
          RvalOut(1);
          RvalOut(2);
       }
       if (ii == 1) {
          RvalOut(0);
          RvalOut(2);
       }
       if (ii == 2) {
          RvalOut(0);
          RvalOut(1);
       }
    }
    goto end;

  } // end (PartFound == PART_RESISTOR)
#endif

#ifdef C_MESS
//Kapazittsmessung auch nur auf Mega8 verfgbar
  else if(PartFound == PART_CAPACITOR) {
//     lcd_fix_string(Capacitor);
     lcd_ziff1(ca);			//Pin-Angaben
     lcd_fix_string(CapZeich);
     lcd_ziff1(cb);
     Line2(); 				//2. Zeile
     lcd_show_format_cap();
     goto end;
  }
#endif
  if(NumOfDiodes == 0) { //Keine Dioden gefunden
     lcd_fix_string(TestFailed1); 	//"Kein,unbek. oder"
     Line2(); //2. Zeile
     lcd_fix_string(TestFailed2); 	//"defektes "
     lcd_fix_string(Bauteil);		//"Bauteil"
  } else {
     lcd_fix_string(Bauteil);		//"Bauteil"
     lcd_fix_string(Unknown); 		//" unbek."
     Line2(); //2. Zeile
     lcd_fix_string(OrBroken); 		//"oder defekt "
     lcd_data(NumOfDiodes + '0');
     lcd_fix_string(AnKat);		//"->|-"
  }
  goto end;


gakAusgabe:
  Line2(); //2. Zeile
  lcd_fix_string(GAK);		//"GAK="
  lcd_ziff1(trans.b);
  lcd_ziff1(trans.c);
  lcd_ziff1(trans.e);
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - 
 end:
  while(!(ON_PIN_REG & (1<<RST_PIN)));	//warten ,bis Taster losgelassen
  wait200ms();
// 10 Sekunden warten
  for(gthvoltage = 0;gthvoltage<10000;gthvoltage++) {
     if(!(ON_PIN_REG & (1<<RST_PIN))) {
        // Wenn der Taster wieder gedrckt wurde...
        // wieder zum Anfang springen und neuen Test durchfhren
        goto start;
     }
     wdt_reset();
     wait1ms();
  }
  wdt_disable();			//Watchdog aus
  ON_PORT &= ~(1<<ON_PIN);		//Abschalten
  //Endlosschleife
  while(1) {
     if(!(ON_PIN_REG & (1<<RST_PIN))) {
        // wird nur erreicht,
        // wenn die automatische Abschaltung nicht eingebaut wurde 
        goto start;
     }
  }
  return 0;
}   // end main


//******************************************************************
// Ausgabe der Flussspannung fr 1-2 Dioden in Zeile 2
// bcdnum = Nummern der beiden Dioden:
// oberer 4 Bit  Index der ersten Diode
// untere 4 Bit  Index der zweiten Diode (Struktur diodes[nn])
// wenn Index >= 3 erfolgt keine Ausgabe
void UfAusgabe(uint8_t bcdnum) {

   Line2(); 				//2. Zeile
   lcd_fix_string(Uf);			//"Uf="
   mVAusgabe(bcdnum >> 4);
   mVAusgabe(bcdnum & 0x0f);
}
void mVAusgabe(uint8_t nn) {
   if (nn < 3) {
      lcd_string(utoa(diodes[nn].Voltage, outval, 10));
      lcd_data('m');
      lcd_data('V');
      lcd_data(' ');
   }
}

#ifdef R_MESS
void RvalOut(uint8_t ii) {	
   // Widerstandswert ausgeben
   uint8_t strlength;
   unsigned long rrx;
   char prefix;
   uint8_t jj,kk;

   rrx = resis[ii].rx;
   prefix = (char)0;
   if (rrx > 999999) {
      rrx = (rrx + 500) / 1000;
      prefix = 'k';
   }
   jj = 0;
   if (rrx > 99999) {
      rrx = (rrx + 50) / 100;
      jj  = 2;
      if (prefix == (char)0) {
         prefix = 'k';
      } else {
         prefix = 'M';
      }
   }
   if (rrx > 9999) {
      rrx = (rrx + 5) / 10;
      jj = 3;
      if (prefix == (char)0) {
         prefix = 'k';
      } else {
         prefix = 'M';
      }
   }
   ultoa(rrx,outval,10);
   strlength = strlen(outval);	//Ntig, um Komma anzuzeigen
   for (kk=0;kk<strlength;kk++) {
      lcd_data(outval[kk]);
      if ((kk + jj) == strlength) lcd_data('.');
   }

   if (prefix != (char)0) {
      lcd_data(prefix);
   }
   lcd_data(LCD_CHAR_OMEGA);	//Omega fr Ohm 
   lcd_data(' ');
 }
#endif

//******************************************************************
//#include "ReadADC.c"
#include "CheckPins.c"


void ChargePin10ms(uint8_t PinToCharge, uint8_t ChargeDirection) {
   //Anschluss eines Bauelementes kurz(10ms) auf ein bestimmtes Potenzial legen
   //Diese Funktion ist zum Entladen von MOSFET-Gates vorgesehen, um Schutzdioden u.. in MOSFETs
   //erkennen zu knnen
   //Parameter:
   //PinToCharge: zu entladender Pin als Bits fr den R-Port
   //ChargeDirection: 0 = gegen Masse (N-Kanal-FET), 1= gegen Plus(P-Kanal-FET)

   if(ChargeDirection&1) {
      R_PORT |= PinToCharge;	//R_PORT auf 1 (VCC) 
   } else {
      R_PORT &= ~PinToCharge; // oder 0 (GND)
   }
   R_DDR |= PinToCharge;			//Pin auf Ausgang, ber R auf Masse oder VCC
   wait10ms();
   // hochohmig schalten, Input
   R_DDR &= ~PinToCharge;	// R fr Pin wieder auf Eingang
   R_PORT &= ~PinToCharge;	// kein Pullup
}

// erst vorhandene Restladung entladen
void EntladePins() {
  uint8_t adc_gnd;		// Maske der ADC-Ausgnge, die direkt auf GND gelegt werden
  unsigned int adcmv[3];	// Spannungswerte der 3 Pins in mV
  unsigned int clr_cnt;		// Clear Counter
  uint8_t lop_cnt;
// Entladezeit in ms / Wartezeit (10000/20)
#define MAX_ENTLADE_ZEIT  (10000/20)

  for(lop_cnt=0;lop_cnt<10;lop_cnt++) {
     adc_gnd = TXD_MSK;		// alle ADC auf Input
     ADC_DDR = adc_gnd;
     ADC_PORT = TXD_VAL;		// ADC-Ausgnge auf 0
     R_PORT = 0;			// R-Ausgnge auf 0
     R_DDR = (2<<(PC2*2)) | (2<<(PC1*2)) | (2<<(PC0*2)); // alle Pins ber R_H auf Masse
     adcmv[0] = W5msReadADC(PC0);	// welche Spannung hat Pin 1?
     adcmv[1] = ReadADC(PC1);	// welche Spannung hat Pin 2?
     adcmv[2] = ReadADC(PC2);	// welche Spannung hat Pin 3?
     if ((adcmv[0] < 3) & (adcmv[1] < 3) & (adcmv[2] < 3)) {
        ADC_DDR = TXD_MSK;		// alle direkten ADC-Ausgnge hochohmig
        R_DDR = 0;			// alle R_L Ports (und R_H) hochohmig
        return;
     }
     // alle Pins mit weniger als 1V direkt ber ADC-Port auf Masse (GND) schalten
     if (adcmv[0] < 1000) {
        adc_gnd |= (1<<PC0);	//Pin 1 ber Ausgang des Analog-Pins auf GND
     }
     if (adcmv[1] < 1000) {
        adc_gnd |= (1<<PC1);	//Pin 2 ber Ausgang des Analog-Pins auf GND
     }
     if (adcmv[2] < 1000) {
        adc_gnd |= (1<<PC2);	//Pin 3 ber Ausgang des Analog-Pins auf GND
     }
     ADC_DDR = adc_gnd;		// jetzt alle gewhlten ADC-Ports auf GND schalten

     // zustzlich die verbleibenden Ports ber R_L auf Masse.
     // Da es bei den direkt auf Masse geschalteten Pins auch nicht schadet, werden
     // einfach alle R_L Widerstnde geschaltet
     R_DDR = (1<<(PC2*2)) | (1<<(PC1*2)) | (1<<(PC0*2));	// Pins ber R_L Widerstnde auf GND
     for(clr_cnt=0;clr_cnt<MAX_ENTLADE_ZEIT;clr_cnt++) {
        wdt_reset();
        adcmv[0] = W20msReadADC(PC0);	// welche Spannung hat Pin 1?
        adcmv[1] = ReadADC(PC1);	// welche Spannung hat Pin 2?
        adcmv[2] = ReadADC(PC2);	// welche Spannung hat Pin 3?
        if (adcmv[0] < 1300) {
           ADC_DDR |= (1<<PC0);	// unter 1.3V ber ADC-Port auf GND
        }
        if (adcmv[1] < 1300) {
           ADC_DDR |= (1<<PC1);	// unter 1.3V ber ADC-Port auf GND
        }
        if (adcmv[2] < 1300) {
           ADC_DDR |= (1<<PC2);	// unter 1.3V ber ADC-Port auf GND
        }
        if ((adcmv[0] < 5) && (adcmv[1] < 5) && (adcmv[2] < 5)) {
           break;
        }
     }
     if (clr_cnt == MAX_ENTLADE_ZEIT) {
        // Kondensator hat Restladung, spter warnen!
//        Line3();
//        lcd_data('E');
//        lcd_string(utoa(adcmv[0], outval, 10));
//        lcd_data(' ');
//        lcd_string(utoa(adcmv[1], outval, 10));
//        lcd_data(' ');
//        lcd_string(utoa(adcmv[2], outval, 10));
     }
     R_PORT = 0;		// alle R_Ausgnge auf 0
     for(adcmv[0]=0;adcmv[0]<clr_cnt;adcmv[0]++) {
        // Zur Sicherheit nochmal 1/20 der Zeit im Kurzschluss-Mode
        wait1ms();
     }
  } // end while
 }




#ifdef C_MESS	//Kapazittsmessung nur auf Mega8 verfgbar
#include "ReadCapacity.c"


void lcd_show_format_cap(void) {
  uint8_t strlength;
  while(cval > 399999) {	//ab 400nF oder 400F
//   cval /= 1000;
     cval = (cval + 500) / 1000;	// mit Rundung
     cpre++;				// F-prefix um 3 Zehnerpotenzen erhhen
  }
  ultoa(cval, outval, 10);
  strlength = strlen(outval);
  if (strlength == 6) { // letzten 2 Stellen weglassen
     outval[4] = outval[3];		// Platz schaffen fr '.'
     outval[3] = '.';
     cpre++;				// F-prefix um 3 Zehnerpotenzen erhhen
  }
  if (strlength == 5) {
     outval[4] = outval[3];		// Platz schaffen fr '.'
     outval[3] = outval[2];
     outval[2] = '.';
     cpre++;				// F-prefix um 3 Zehnerpotenzen erhhen
  }
  outval[5] = 0;

  lcd_string(outval);
  tmpval2 = pgm_read_byte(&C_Prefix_tab[cpre]);
  lcd_data(tmpval2);		// p oder n oder  oder m
  lcd_data('F');
 }
unsigned int getRLmultip(unsigned int cvolt) {

// Widerstand 680 Ohm         250  275  300  325  350  375  400  425  450  475  500  525  550  575  600  625  650  675  700  725  750  775  800  825  850  875  900  925  950  975 1000 1025 1050 1075 1100 1125 1150 1175 1200 1225 1250 1275 1300 1325 1350 1375 1400  mV
//uint16_t RLtab[] MEM_TEXT = {27077,24552,22447,20665,19138,17815,16657,15635,14727,13914,13182,12520,11918,11369,10865,10401,9973,9577,9209,8866,8546,8247,7966,7702,7454,7220,6999,6789,6591,6403,6224,6054,5892,5738,5590,5449,5314,5185,5061,4942,4828,4718,4613,4511,4413,4319,4228};

#define RL_Tab_Abstand 25		// Tabellenabstand ist 50mV
#define RL_Tab_Beginn 250	// Tabellenbeginn ist 250mV
  unsigned int uvolt;
  unsigned int y1, y2;
  uint8_t tabind;
  uint8_t tabres;
  if (cvolt >= RL_Tab_Beginn) {
     uvolt = cvolt - RL_Tab_Beginn;
  } else {
     uvolt = 0;		// begrenze auf Tabellenbeginn
  }
  tabind = uvolt / RL_Tab_Abstand;
  tabres = uvolt % RL_Tab_Abstand;
  tabres = RL_Tab_Abstand - tabres;
  if (tabind > 45) {
     tabind = 45;
  }
  #ifdef USE_EEPROM
  y1 = eeprom_read_word(&RLtab[tabind]);
  y2 = eeprom_read_word(&RLtab[tabind+1]);
  #else
  y1 = pgm_read_word(&RLtab[tabind]);
  y2 = pgm_read_word(&RLtab[tabind+1]);
  #endif
  return ( ((y1 - y2) * tabres + (RL_Tab_Abstand/2)) / RL_Tab_Abstand + y2);
}


 #ifdef WITH_AUTO_REF
void RefVoltage(void) {
// zur Erinnerung:
//Widerstand 470000 Ohm      1000 1050 1100 1150 1200 1250 1300 1350 1400  mV
//uint16_t RHTAB[] MEM_TEXT = { 954, 903, 856, 814, 775, 740, 707, 676, 648};
#define Ref_Tab_Abstand 50		// Tabellenabstand ist 50mV
#define Ref_Tab_Beginn 1000		// Tabelle beginnt bei 1000mV

  unsigned int referenz;
  unsigned int y1, y2;
  uint8_t tabind;
  uint8_t tabres;

  referenz = ref_mv;
  if (referenz >= Ref_Tab_Beginn) {
     referenz -= Ref_Tab_Beginn;
  } else  {
     referenz = 0;		// Begrenze auf Tabellenbeginn
  }
  tabind = referenz / Ref_Tab_Abstand;
  tabres = referenz % Ref_Tab_Abstand;
  tabres = Ref_Tab_Abstand-tabres;
  if (tabind > 7) {
     tabind = 7;
  }
  // Interpoliere die Tabelle der Faktoren
  #ifdef USE_EEPROM
  y1 = eeprom_read_word(&RHtab[tabind]);
  y2 = eeprom_read_word(&RHtab[tabind+1]);
  #else
  y1 = pgm_read_word(&RHtab[tabind]);
  y2 = pgm_read_word(&RHtab[tabind+1]);
  #endif
  // RHmultip ist der interpolierte Faktor fr 470k
  RHmultip = ((y1 - y2) * tabres + (Ref_Tab_Abstand/2)) / Ref_Tab_Abstand + y2;
 }
 #endif
#endif

#ifdef LCD_CLEAR
void lcd_clear_line(void) {
 // Schreibt 20 Leerzeichen auf Display, Cursor muss am Anfang der Zeile stehen
 unsigned char ll;
 for (ll=0;ll<20;ll++) {
    lcd_data(' ');
 }
}
#endif

#ifdef WITH_SELFTEST
void AutoCheck(void) {
  uint8_t tt;		// laufende Testnummer
  uint8_t ww;		// Zhler fr Testwiederholung
  unsigned int  adcmv[3];
  ADC_PORT = TXD_VAL;
  ADC_DDR = TXD_MSK;
  R_PORT = (1<<(TP1*2));		//Pin 1 ber RL auf +
  R_DDR = (1<<(TP1*2)) | (1<<(TP2*2));	//Pin 2 ber RL auf -
  adcmv[0] = W5msReadADC(PC0);		// Spannung an Pin 1
  adcmv[1] = ReadADC(PC1);		// Spannung an Pin 2
  adcmv[2] = ReadADC(PC2);		// Spannung an Pin 3
  if (abs((int)(adcmv[1] - adcmv[0])) > 20) {
     return;				//Abweichung zu gross, kein Selbsttest
  }
  if (abs((int)(adcmv[2] - adcmv[0])) > 20) {
     return;				//Abweichung zu gross, kein Selbsttest
  }
  if (abs((int)(adcmv[0] - (U_VCC/2))) > 30) {
     return;				//Abweichung zu gross, kein Selbsttest
  }
  R_DDR = (1<<(TP1*2)) | (1<<(TP3*2));	//Pin 3 ber RL auf - (Pin 1 ber RL auf +)
  adcmv[0] = W5msReadADC(PC0);		// Spannung an Pin 1
  adcmv[1] = ReadADC(PC1);		// Spannung an Pin 2
  adcmv[2] = ReadADC(PC2);		// Spannung an Pin 3
  if (abs((int)(adcmv[1] - adcmv[0])) > 20) {
     return;				//Abweichung zu gross, kein Selbsttest
  }
  if (abs((int)(adcmv[2] - adcmv[0])) > 20) {
     return;				//Abweichung zu gross, kein Selbsttest
  }
  if (abs((int)(adcmv[0] - (U_VCC/2))) > 30) {
     return;				//Abweichung zu gross, kein Selbsttest
  }
  lcd_clear();
  Line1();
  lcd_fix_string(SELFTEST);		// "Selftest mode.."
  wait1s();
 
  for(tt=1;tt<10;tt++) {			// Scheife ber Tests
     for(ww=0;ww<8;ww++) {		// Testwiederholung
        Line2();			//Cursor auf Spalte1, Zeile2
        lcd_clear_line();
        Line1();			//Cursor auf Spalte1, Zeile1
        lcd_clear_line();
        Line1();			//Cursor auf Spalte1, Zeile1
        lcd_data('T');
        lcd_data('0'+tt);
        lcd_data(' ');
        if (tt == 1) {   // Ausgabe der Referenzspannung und Faktoren fr C-Messung
           ref_mv = ReadADC(0x0e);      // Referenz-Spannung lesen
           ref_mv = W5msReadADC(0x0e) + REF_KORR;  // Referenz-Spannung lesen
           lcd_fix_string(URefT);	//"URef="
           lcd_string(utoa(ref_mv, outval, 10));
           lcd_fix_string(mVT);		//"mV "
           Line2();
           lcd_fix_string(RHfakt);	//"RHf="
           lcd_string(utoa(RHmultip, outval, 10));
#ifdef C_MESS
           lcd_fix_string(RLfakt);	//" RLf="
           lcd_string(utoa(getRLmultip(ref_mv), outval, 10));
#endif
        }
        if (tt == 2) { // Messung gleiche Widerstnde RL
           R_PORT = 1<<(TP1*2);		//RL1 auf VCC
           R_DDR = (1<<(TP1*2)) | (1<<(TP2*2));	//RL2 auf -
           adcmv[0] = W20msReadADC(TP1);
           R_DDR = (1<<(TP1*2)) | (1<<(TP3*2));	//RL3 auf -
           adcmv[1] = W20msReadADC(TP1);
           R_PORT = 1<<(TP2*2);		//RL2 auf VCC
           R_DDR = (1<<(TP2*2)) | (1<<(TP3*2));	//RL3 auf -
           adcmv[2] = W20msReadADC(TP1);
           lcd_fix_string(RLRL);	// "RLRL"
        }
        if (tt == 3) { // Messung gleiche Widerstnde RH
           R_PORT = 2<<(TP1*2);		//RH1 auf VCC
           R_DDR = (2<<(TP1*2)) | (2<<(TP2*2));	//RH2 auf -
           adcmv[0] = W20msReadADC(TP1);
           R_DDR = (2<<(TP1*2)) | (2<<(TP3*2));	//RH3 auf -
           adcmv[1] = W20msReadADC(TP1);
           R_PORT = 2<<(TP2*2);		//RL2 auf VCC
           R_DDR = (2<<(TP2*2)) | (2<<(TP3*2));	//RH3 auf -
           adcmv[2] = W20msReadADC(TP1);
           lcd_fix_string(RHRH);	// "RHRH"
        }
        if (tt == 4) { // Text trennen
           lcd_fix_string(RELPROBE);	// "Release Probes"
        }
        if (tt == 5) { // Messung, wie gut R_H die Spannung auf Ground und VCC zieht
           R_PORT = 0;
           R_DDR = 2<<(TP1*2);		//Pin 1 ber R_H auf Masse
           adcmv[0] = W20msReadADC(TP1);

           R_DDR = 2<<(TP2*2);		//Pin 2 ber R_H auf Masse
           adcmv[1] = W20msReadADC(TP2);

           R_DDR = 2<<(TP3*2);		//Pin 3 ber R_H auf Masse
           adcmv[2] = W20msReadADC(TP3);
           lcd_fix_string(RH1L);	// "RH_Lo="
        }
        if (tt == 6) { // Messung, wie gut R_H die Spannung auf VCC zieht
           R_DDR = 2<<(TP1*2);		//Pin 1 ber R_H auf VCC
           R_PORT = 2<<(TP1*2);
           adcmv[0] = W20msReadADC(TP1);
           R_DDR = 2<<(TP2*2);		//Pin 2 ber R_H auf VCC
           R_PORT = 2<<(TP2*2);
           adcmv[1] = W20msReadADC(TP2);
           R_DDR = 2<<(TP3*2);		//Pin 3 ber R_H auf VCC
           R_PORT = 2<<(TP3*2);
           adcmv[2] = W20msReadADC(TP3);
           lcd_fix_string(RH1H);	// "RH_Hi="
        }
        if (tt == 7) { // Messung des Innenwiderstandes der Ports
           ADC_DDR = 1<<TP1 | TXD_MSK;	//alle ADC-Pins auf Ausgang 0V
           R_PORT = 1<<(TP1*2); // Alle R_L-PORTS auf VCC
           R_DDR = 1<<(TP1*2);		//Low-Pin auf Ausgang und ber R_L auf VCC
           adcmv[0] = W5msReadADC(TP1);

           ADC_DDR = 1<<TP2 | TXD_MSK;	//alle ADC-Pins auf Ausgang 0V
           R_PORT =  1<<(TP2*2); // Alle R_L-PORTS auf VCC
           R_DDR = 1<<(TP2*2);		//High-Pin auf Ausgang und ber R_L auf VCC
           adcmv[1] = W5msReadADC(TP2);

           ADC_DDR = 1<<TP3 | TXD_MSK;	//alle ADC-Pins auf Ausgang 0V
           R_PORT =  1<<(TP3*2); // Alle R_L-PORTS auf VCC
           R_DDR = 1<<(TP3*2);		//High-Pin auf Ausgang und ber R_L auf VCC
           adcmv[2] = W5msReadADC(TP3);
           lcd_fix_string(RILO);	// "RiLo="
        }
        if (tt == 8) { // Messung des Innenwiderstandes der Ports
           R_PORT = 0;			// alle R-Ports auf Masse
           ADC_PORT = 1<<TP1 | TXD_VAL;	// alle ADC-Ports auf VCC
           ADC_DDR = 1<<TP1 | TXD_MSK;	//alle ADC-Pins auf Ausgang 0V
           R_DDR = 1<<(TP1*2);			//Low-Pin auf Ausgang und ber R_L auf Masse
           adcmv[0] = U_VCC - W5msReadADC(TP1);
      
           ADC_PORT = 1<<TP2 | TXD_VAL;	// alle ADC-Ports auf VCC
           ADC_DDR = 1<<TP2 | TXD_MSK;	//alle ADC-Pins auf Ausgang 0V
           R_DDR = 1<<(TP2*2);		//High-Pin auf Ausgang und ber R_L auf Masse
           adcmv[1] = U_VCC - W5msReadADC(TP2);

           ADC_PORT = 1<<TP3 | TXD_VAL;	// alle ADC-Ports auf VCC
           ADC_DDR = 1<<TP3 | TXD_MSK;	//alle ADC-Pins auf Ausgang 0V
           R_DDR = 1<<(TP3*2);		//High-Pin auf Ausgang und ber R_L auf Masse
           adcmv[2] = U_VCC - W5msReadADC(TP3);

           lcd_fix_string(RIHI);	// "RiHi="
        }
        if (tt == 9) { 			//Frequenz-Generator 50Hz
           ADC_PORT = TXD_VAL;
           ADC_DDR = 1<<TP1 | TXD_MSK;	// Pin 1 auf GND
           R_DDR = (1<<(TP3*2)) | (1<<(TP2*2));
           for (ii=0;ii<250;ii++) {
              R_PORT = (1<<(TP2*2));	// Pin 2 ber R_L auf VCC, Pin 3 ber R_L auf GND
              wait10ms();
              R_PORT = (1<<(TP3*2));	// Pin 3 ber R_L auf VCC, Pin 2 ber R_L auf GND
              wait10ms();
              wdt_reset();
           }
        }
        if (tt > 1) {			// die gemessenen 3 Spannungen ausgeben
           Line2();			//Cursor auf Spalte1, Zeile2
           lcd_string(utoa(adcmv[0], outval, 10));
           lcd_data(' ');
           lcd_string(utoa(adcmv[1], outval, 10));
           lcd_data(' ');
           lcd_string(utoa(adcmv[2], outval, 10));
        }
        ADC_DDR =  TXD_MSK;		//alle-Pins auf Input
        ADC_PORT = TXD_VAL;		// alle ADC-Ports auf GND
        R_DDR = 0;			// alle R-Ports auf Input
        R_PORT = 0;
        if(!(ON_PIN_REG & (1<<RST_PIN))) {
           break;
        }
        wait500ms();
        if(!(ON_PIN_REG & (1<<RST_PIN))) {
           break;
        }
        wait500ms();
     } //end for ww
     wait1s();
  } //end for tt
  lcd_clear();
  Line1();
  lcd_fix_string(ATE);		//"Auto Test End"
  Line2();
  lcd_fix_string(VERSION);	//"Version ..."
  wait2s();			//warte 2 Sekunden
 } 
#endif
